"""Calculate functions, part of gptools.
"""

import sys
import math

import utils.exceptions as exceptions
import convert.units as units
from utils.constants import ellipsoids

def distanceVicenty(lat1, long1, lat2, long2, ellipsoid='WGS-84'):
    """Computes the Vicenty distance (in meters) between two points
    on the earth. Coordinates need to be in decimal degrees.
    """
    # Check if we got numbers
    if not ((type(lat1) != type(float()) or type(lat1) != type(int())) and (type(long1) != type(float()) or type(long1) != type(int())) and (type(lat2) != type(float()) or type(lat2) != type(int())) and (type(long2) != type(float()) or type(long2) != type(int()))): 
        raise exceptions.NotNumericValue("Values are not numeric. Cannot compute distance.")
    # Check if we know about the ellipsoid
    if ellipsoid not in ellipsoids.keys():
        sys.exit("Unknown ellipsoid " + ellipsoid)
    else:
        major, minor, ffactor = ellipsoids[ellipsoid]
    # Convert degrees to radians
    x1 = units.degreesToRadians(lat1)
    y1 = units.degreesToRadians(long1)
    x2 = units.degreesToRadians(lat2)
    y2 = units.degreesToRadians(long2)
    # Define our flattening f
    f = 1 / ffactor
    # Find delt X
    deltaX = y2 - y1
    # Calculate U1 and U2
    U1 = math.atan((1 - f) * math.tan(x1))
    U2 = math.atan((1 - f) * math.tan(x2))
    # Calculate the sin and cos of U1 and U2
    sinU1 = math.sin(U1)
    cosU1 = math.cos(U1)
    sinU2 = math.sin(U2)
    cosU2 = math.cos(U2)
    # Set initial value of L
    L = deltaX
    # Set Lambda equal to L
    lmbda = L
    # Iteration limit - when to stop if no convergence
    iterLimit = 100
    while abs(lmbda) > 10e-12 and iterLimit >= 0:
        # Calculate sine and cosine of lmbda
        sin_lmbda = math.sin(lmbda)
        cos_lmbda = math.cos(lmbda)
        # Calculate the sine of sigma
        sin_sigma = math.sqrt(
                (cosU2 * sin_lmbda) ** 2 + 
                (cosU1 * sinU2 - 
                 sinU1 * cosU2 * cos_lmbda) ** 2
        )
        if sin_sigma == 0.0:
            # Concident points - distance is 0
            return 0.0
        # Calculate the cosine of sigma
        cos_sigma = (
                    sinU1 * sinU2 + 
                    cosU1 * cosU2 * cos_lmbda
        )
        # Calculate sigma
        sigma = math.atan2(sin_sigma, cos_sigma)
        # Calculate the sine of alpha
        sin_alpha = (cosU1 * cosU2 * math.sin(lmbda)) / (sin_sigma)
        # Calculate the square cosine of alpha
        cos_alpha_sq = 1 - sin_alpha ** 2
        # Calculate the cosine of 2 sigma
        cos_2sigma = cos_sigma - ((2 * sinU1 * sinU2) / cos_alpha_sq)
        # Identify C
        C = (f / 16.0) * cos_alpha_sq * (4.0 + f * (4.0 - 3 * cos_alpha_sq))
        # Recalculate lmbda now
        lmbda = L + ((1.0 - C) * f * sin_alpha * (sigma + C * sin_sigma * (cos_2sigma + C * cos_sigma * (-1.0 + 2 * cos_2sigma ** 2)))) 
        # If lambda is greater than pi, there is no solution
        if (abs(lmbda) > math.pi):
            raise ValueError("No solution can be found.")
        iterLimit -= 1
    if iterLimit == 0 and lmbda > 10e-12:
        raise ValueError("Solution could not converge.")
    # Since we converged, now we can calculate distance
    # Calculate u squared
    u_sq = cos_alpha_sq * ((major ** 2 - minor ** 2) / (minor ** 2))
    # Calculate A
    A = 1 + (u_sq / 16384.0) * (4096.0 + u_sq * (-768.0 + u_sq * (320.0 - 175.0 * u_sq)))
    # Calculate B
    B = (u_sq / 1024.0) * (256.0 + u_sq * (-128.0 + u_sq * (74.0 - 47.0 * u_sq)))
    # Calculate delta sigma
    deltaSigma = B * sin_sigma * (cos_2sigma + 0.25 * B * (cos_sigma * (-1.0 + 2.0 * cos_2sigma ** 2) - 1.0/6.0 * B * cos_2sigma * (-3.0 + 4.0 * sin_sigma ** 2) * (-3.0 + 4.0 * cos_2sigma ** 2)))
    # Calculate s, the distance
    s = minor * A * (sigma - deltaSigma)
    # Return the distance
    return s


def distanceGreatCircle(lat1, long1, lat2, long2, radius=6371009.0):
    """Computes the great circle distance (in meters) between two 
    points in decimal degrees. Mean Earth Radius used is IUGG.
    """
    # Check if we got numbers
    if not ((type(lat1) != type(float()) or type(lat1) != type(int())) and (type(long1) != type(float()) or type(long1) != type(int())) and (type(lat2) != type(float()) or type(lat2) != type(int())) and (type(long2) != type(float()) or type(long2) != type(int()))): 
        raise exceptions.NotNumericValue("Values are not numeric. Cannot compute distance.")
    # Convert degrees to radians
    x1 = units.degreesToRadians(lat1)
    y1 = units.degreesToRadians(long1)
    x2 = units.degreesToRadians(lat2)
    y2 = units.degreesToRadians(long2)
    # Calculate the great circle distance
    # Helper variables so the equation looks better
    sin_x1 = math.sin(x1)
    cos_x1 = math.cos(x1)
    sin_x2 = math.sin(x2)
    cos_x2 = math.cos(x2)
    dlong = float(y2) - float(y1)
    sin_dlong = math.sin(dlong) 
    cos_dlong = math.cos(dlong)
    # Delta sigma equation
    delta_sigma = math.atan2(
            math.sqrt( ((cos_x2 * sin_dlong) ** 2) +
            ((cos_x1 * sin_x2 - sin_x1 * cos_x2 * cos_dlong) ** 2)
        ) , (
            (sin_x1 * sin_x2) + 
            (cos_x1 * cos_x2 * cos_dlong)
        )
    )
    print delta_sigma
    return radius * delta_sigma           
    