"""Feature Class conversion methods, part of gptools.
"""

# Import all necessary modules
import os                  # Could be security risk since file is not locked
import sys
from xml.dom import minidom

import gptools.utils as utils
import gptools.utils.exceptions as exceptions

def pointsToLines(pointFile, outputFile, whereClause=None, lineDiff=None):
    """
    pointsToLines script, part of gptools. Creates a line feature class from a point
    feature class. Optionally, differentiates lines based on a field. No variables
    are saved along, other than geometry and line ID, if asking for differentiation.
    """
    # Begin checking conditions before we load the geoprocessor
    # Does the input file exist?
    if os.path.isfile(pointFile):
        # Is the input file indeed a point feature class?
        # Load the geoprocessor and check
        gp = utils.getGP()
        desc = gp.Describe(pointFile)
        # Check if the DataType is of ShapeFile or Featureclass
        if desc.ShapeType not in utils.constants.point:
            raise exceptions.WrongGeometryType("The expected input geometry type was point, not " + desc.DataType)
        # Does the output file exist? And can we overwrite it?
        if (os.path.isfile(outputFile) and gp.OverwriteOutput) or not os.path.isfile(outputFile):
            # We can write to outputFile safely. Start reading the input
            if whereClause == None:
                whereClause = ""
            if lineDiff != None:
                # Did they have a line differantiator? Use it if they did
                pointFileFields = [x.Name for x in desc.Fields]
                if lineDiff not in pointFileFields:
                    raise exceptions.FieldNotExisting("The field " + str(lineDiff) + " does not exist.")
                else:
                    lineDiffSC = str(lineDiff)
            else:
                lineDiffSC = ""
            # Grab a search cursor
            try:
                sCur = gp.SearchCursor(pointFile, str(whereClause), desc.SpatialReference, "", lineDiffSC)
            except:
                sys.exit("Error opening a SearchCursor on file " + str(pointFile) + ". " + str(sys.exc_info()))
            # Create the output file
            gp.CreateFeatureClass(os.path.dirname(outputFile), os.path.basename(outputFile), "POLYLINE", "#", "#", "#", desc.SpatialReference)
            descOut = gp.Describe(outputFile)
            # If we add a differentiator, add a field for it.
            if lineDiff != None:
                # Find the field type
                for field in desc.Fields:
                    if field.Name == lineDiff:
                        diffType = str(field.Type).upper()
                        if diffType == "SMALLINTEGER":
                            diffType = "SHORT"
                        diffPrec = field.Precision
                        diffScale = field.Scale
                gp.AddField(outputFile, str(lineDiff), diffType, diffPrec, diffScale)
            # Grab an insert cursor
            iCur = gp.InsertCursor(outputFile, desc.SpatialReference)
            # Start reading the file
            sRow = sCur.Next()
            # Create an array
            lineArray = gp.CreateObject("Array")
            # If we have a line differentiator, go on this way.
            if lineDiff != None:
                # Get the first line diff value
                curDiff = sRow.GetValue(lineDiff)
                # start looping through rows
                while sRow:
                    # if the row has the same diff value as the previous one, add it
                    if sRow.GetValue(lineDiff) == curDiff:
                        pnt = sRow.GetValue(desc.ShapeFieldName)
                        lineArray.add(pnt.GetPart())
                    else:
                        # We have a new line here. Add the old one
                        # Create row
                        iRow = iCur.NewRow()
                        # Set row's shape to lineArray
                        iRow.SetValue(descOut.ShapeFieldName, lineArray)
                        iRow.SetValue(lineDiff, curDiff)
                        # Insert the row
                        iCur.InsertRow(iRow)
                        # Clear the lineArray
                        lineArray.RemoveAll()
                        # Delete the iRow from memory
                        del(iRow)
                        # Get the new differentiator
                        curDiff = sRow.GetValue(lineDiff)
                        # Grab the point
                        pnt = sRow.GetValue(desc.ShapeFieldName)
                        # Add it to the array
                        lineArray.add(pnt.GetPart())
                    sRow = sCur.Next()
                # Add the final array
                iRow = iCur.NewRow()
                iRow.SetValue(descOut.ShapeFieldName, lineArray)
                iRow.SetValue(lineDiff, curDiff)
                iCur.InsertRow(iRow)
            else:
                # If there is no diff value, just go.
                while sRow:
                    # Get the point
                    pnt = sRow.GetValue(desc.ShapeFieldName)
                    # Add it to the array
                    lineArray.add(pnt.GetPart())
                    # Grab next row
                    sRow = sCur.Next()
                # Create the line feature here
                iRow = iCur.NewRow()
                # Set row's shape to lineArray
                iRow.SetValue(descOut.ShapeFieldName, lineArray)
                # Insert the row
                iCur.InsertRow(iRow)
                # Clear the lineArray
                lineArray.RemoveAll()
        else:
            # File exists but OverwriteOutput is not set. We need to exit gracefully.
            raise IOError("File " + str(outputFile) + " already exists and OverwriteOutput is set to false.")
    else:
        raise IOError("File " + str(pointFile) + " does not exist.")
    del iCur
    del sCur
    # Return
    return outputFile

def pointsToPolygons(pointFile, outputFile, whereClause=None, polygonDiff=None):
    """
    pointsToPolygons script, part of gptools. Creates a polygon feature class from a point
    feature class. Optionally, differentiates polygonss based on a field. No variables
    are saved along, other than geometry and polygon ID, if asking for differentiation.
    """
    # Begin checking conditions before we load the geoprocessor
    # Does the input file exist?
    if os.path.isfile(pointFile):
        # Is the input file indeed a point feature class?
        # Load the geoprocessor and check
        gp = utils.getGP()
        desc = gp.Describe(pointFile)
        # Check if the DataType is of ShapeFile or Featureclass
        if desc.ShapeType not in utils.constants.point:
            raise exceptions.WrongGeometryType("The expected input geometry type was point, not " + desc.DataType)
        # Does the output file exist? And can we overwrite it?
        if (os.path.isfile(outputFile) and gp.OverwriteOutput) or not os.path.isfile(outputFile):
            # We can write to outputFile safely. Start reading the input
            if whereClause == None:
                whereClause = ""
            if polygonDiff != None:
                # Did they have a line differantiator? Use it if they did
                pointFileFields = [x.Name for x in desc.Fields]
                if polygonDiff not in pointFileFields:
                    raise exceptions.FieldNotExisting("The field " + str(polygonDiff) + " does not exist.")
                else:
                    polygonDiffSC = str(polygonDiff)
            else:
                polygonDiffSC = ""
            # Grab a search cursor
            try:
                sCur = gp.SearchCursor(pointFile, str(whereClause), desc.SpatialReference, "", polygonDiffSC)
            except:
                sys.exit("Error opening a SearchCursor on file " + str(pointFile) + ". " + str(sys.exc_info()))
            # Create the output file
            gp.CreateFeatureClass(os.path.dirname(outputFile), os.path.basename(outputFile), "POLYGON", "#", "#", "#", desc.SpatialReference)
            descOut = gp.Describe(outputFile)
            # If we add a differentiator, add a field for it.
            if polygonDiff != None:
                # Find the field type
                for field in desc.Fields:
                    if field.Name == polygonDiff:
                        diffType = str(field.Type).upper()
                        if diffType == "SMALLINTEGER":
                            diffType = "SHORT"
                        diffPrec = field.Precision
                        diffScale = field.Scale
                gp.AddField(outputFile, str(polygonDiff), diffType, diffPrec, diffScale)
            # Grab an insert cursor
            iCur = gp.InsertCursor(outputFile, desc.SpatialReference)
            # Start reading the file
            sRow = sCur.Next()
            # Create an array
            lineArray = gp.CreateObject("Array")
            # If we have a line differentiator, go on this way.
            if polygonDiff != None:
                # Get the first line diff value
                curDiff = sRow.GetValue(polygonDiff)
                # start looping through rows
                while sRow:
                    # if the row has the same diff value as the previous one, add it
                    if sRow.GetValue(polygonDiff) == curDiff:
                        pnt = sRow.GetValue(desc.ShapeFieldName)
                        lineArray.add(pnt.GetPart())
                    else:
                        # We have a new line here. Add the old one
                        # Create row
                        iRow = iCur.NewRow()
                        # Set row's shape to lineArray
                        iRow.SetValue(descOut.ShapeFieldName, lineArray)
                        iRow.SetValue(polygonDiff, curDiff)
                        # Insert the row
                        iCur.InsertRow(iRow)
                        # Clear the lineArray
                        lineArray.RemoveAll()
                        # Delete the iRow from memory
                        del(iRow)
                        # Get the new differentiator
                        curDiff = sRow.GetValue(polygonDiff)
                        # Grab the point
                        pnt = sRow.GetValue(desc.ShapeFieldName)
                        # Add it to the array
                        lineArray.add(pnt.GetPart())
                    sRow = sCur.Next()
                # Add the final array
                iRow = iCur.NewRow()
                iRow.SetValue(descOut.ShapeFieldName, lineArray)
                iRow.SetValue(polygonDiff, curDiff)
                iCur.InsertRow(iRow)
            else:
                # If there is no diff value, just go.
                while sRow:
                    # Get the point
                    pnt = sRow.GetValue(desc.ShapeFieldName)
                    # Add it to the array
                    lineArray.add(pnt.GetPart())
                    # Grab next row
                    sRow = sCur.Next()
                # Create the line feature here
                iRow = iCur.NewRow()
                # Set row's shape to lineArray
                iRow.SetValue(descOut.ShapeFieldName, lineArray)
                # Insert the row
                iCur.InsertRow(iRow)
                # Clear the lineArray
                lineArray.RemoveAll()
        else:
            # File exists but OverwriteOutput is not set. We need to exit gracefully.
            raise IOError("File " + str(outputFile) + " already exists and OverwriteOutput is set to false.")
    else:
        raise IOError("File " + str(pointFile) + " does not exist.")
    del iCur
    del sCur
    # Return
    return outputFile

def multiPointsToPoints(pointFile, outputFile, whereClause=None):
    """Convert multipoint feature classes to point feature classes.
    """
    # Begin checking conditions before we load the geoprocessor
    # Does the input file exist?
    if os.path.isfile(pointFile):
        # Is the input file indeed a point feature class?
        # Load the geoprocessor and check
        gp = utils.getGP()
        desc = gp.Describe(pointFile)
        # Check if the DataType is of ShapeFile or Featureclass
        if desc.ShapeType not in utils.constants.multipoint:
            raise exceptions.WrongGeometryType("The expected input geometry type was multipoint, not " + desc.ShapeType)
        # Does the output file exist? And can we overwrite it?
        if (os.path.isfile(outputFile) and gp.OverwriteOutput) or not os.path.isfile(outputFile):
            # We can write to outputFile safely. Start reading the input
            if whereClause == None:
                whereClause = ""
            # Find the fields we will need to copy over
            pointFileFields = [x.Name for x in desc.Fields]
            # Grab a search cursor
            try:
                sCur = gp.SearchCursor(pointFile, str(whereClause), desc.SpatialReference, "", "")
            except:
                sys.exit("Error opening a SearchCursor on file " + str(pointFile) + ". " + str(sys.exc_info()))
            # Create the output file
            gp.CreateFeatureClass(os.path.dirname(outputFile), os.path.basename(outputFile), "POINT", pointFile, "", "", desc.SpatialReference)
            descOut = gp.Describe(outputFile)
            # Add a field for the original OID
            gp.AddField(outputFile, "orgID", "LONG")
            # Grab an insert cursor
            iCur = gp.InsertCursor(outputFile, desc.SpatialReference)
            # Start reading the file
            sRow = sCur.Next()
            while sRow:
                # get the first geometry
                mPnt = sRow.GetValue(desc.ShapeFieldName)
                # loop through each part
                for part in range(mPnt.PartCount):
                    # Get the part
                    pnt = mPnt.GetPart(part)
                    # Save the part
                    iRow = iCur.NewRow()
                    # Copy the geometry
                    iRow.SetValue(descOut.ShapeFieldName, pnt)
                    # Copy the fields
                    for field in pointFileFields:
                        if field != desc.OIDFieldName and field != desc.ShapeFieldName:
                            fieldValue = sRow.GetValue(field)
                            iRow.SetValue(field, fieldValue)
                        elif field == desc.OIDFieldName:
                            fieldValue = sRow.GetValue(desc.OIDFieldName)
                            iRow.SetValue("orgID", fieldValue)
                    iCur.InsertRow(iRow)
                    del iRow
                sRow = sCur.Next()
        else:
            # File exists but OverwriteOutput is not set. We need to exit gracefully.
            raise IOError("File " + str(outputFile) + " already exists and OverwriteOutput is set to false.")
    else:
        raise IOError("File " + str(pointFile) + " does not exist.")
    # Return
    del iCur
    del sCur
    return outputFile
    
def pointsToMultiPoints(pointFile, outputFile, whereClause=None, pointDiff=None):
    """Creates a Multi-Point featuce class from a point feature class.
    If a point differentiator is given, it is taken into account, or
    else it creates one multipoint feature. Note that you do net get the
    original fields with it, as no aggregation is performed.
    """
    # Begin checking conditions before we load the geoprocessor
    # Does the input file exist?
    if os.path.isfile(pointFile):
        # Is the input file indeed a point feature class?
        # Load the geoprocessor and check
        gp = utils.getGP()
        desc = gp.Describe(pointFile)
        # Check if the DataType is of ShapeFile or Featureclass
        if desc.ShapeType not in utils.constants.point:
            raise exceptions.WrongGeometryType("The expected input geometry type was point, not " + desc.ShapeType)
        # Does the output file exist? And can we overwrite it?
        if (os.path.isfile(outputFile) and gp.OverwriteOutput) or not os.path.isfile(outputFile):
            # We can write to outputFile safely. Start reading the input
            if whereClause == None:
                whereClause = ""
            if pointDiff == None:
                pointDiff = ""
            # Find the fields we will need to copy over
            pointFileFields = [x.Name for x in desc.Fields]
            # Check if the pntDiff exists
            if pointDiff != "" and pointDiff not in pointFileFields:
                raise ValueError("Point differentiator given does not exist" + str(pointDiff))
            # Grab a search cursor
            try:
                sCur = gp.SearchCursor(pointFile, str(whereClause), desc.SpatialReference, "", str(pointDiff))
            except:
                sys.exit("Error opening a SearchCursor on file " + str(pointFile) + ". " + str(sys.exc_info()))
            # Create the output file
            gp.CreateFeatureClass(os.path.dirname(outputFile), os.path.basename(outputFile), "MULTIPOINT", "", "", "", desc.SpatialReference)
            descOut = gp.Describe(outputFile)
            # Grab an insert cursor
            iCur = gp.InsertCursor(outputFile, desc.SpatialReference)
            mPntArray = gp.CreateObject("Array")
            # Start reading the file
            sRow = sCur.Next()
            if pointDiff != "":
                diff = sRow.GetValue(pointDiff)
            else:
                diff = ""
            while sRow:
                # Check if we are in a new point differentiator
                if pointDiff != "":
                    if diff != sRow.GetValue(pointDiff):
                        diff = sRow.GetValue(pointDiff)
                        # Save the new feature
                        iRow = iCur.NewRow()
                        iRow.SetValue(descOut.ShapeFieldName, mPntArray)
                        iCur.InsertRow(iRow)
                        mPntArray.RemoveAll()
                        del iRow
                # get the first geometry
                pntShape = sRow.GetValue(desc.ShapeFieldName)
                # Get the part
                pnt = pntShape.GetPart(0)
                # Save the part
                mPntArray.add(pnt)
                sRow = sCur.Next()
            iRow = iCur.NewRow()
            # Copy the geometry
            iRow.SetValue(descOut.ShapeFieldName, mPntArray)
            # Insert the row
            iCur.InsertRow(iRow)
            del mPntArray
            del iRow
        else:
            # File exists but OverwriteOutput is not set. We need to exit gracefully.
            raise IOError("File " + str(outputFile) + " already exists and OverwriteOutput is set to false.")
    else:
        raise IOError("File " + str(pointFile) + " does not exist.")
    # Return
    del iCur
    del sCur
    return outputFile

def linesToPoints(lineFile, outputFile, whereClause=None):
    """Converts a line feature class to a point feature class.
    Optionally, uses a where clause to limit what gets converted.
    """
    # Begin checking conditions before we load the geoprocessor
    # Does the input file exist?
    if os.path.isfile(lineFile):
        # Is the input file indeed a point feature class?
        # Load the geoprocessor and check
        gp = utils.getGP()
        desc = gp.Describe(lineFile)
        # Check if the DataType is of ShapeFile or Featureclass
        if desc.ShapeType not in utils.constants.polyline:
            raise exceptions.WrongGeometryType("The expected input geometry type was line, not " + desc.ShapeType)
        # Does the output file exist? And can we overwrite it?
        if (os.path.isfile(outputFile) and gp.OverwriteOutput) or not os.path.isfile(outputFile):
            # We can write to outputFile safely. Start reading the input
            if whereClause == None:
                whereClause = ""
            # Find the fields we will need to copy over
            lineFileFields = [x.Name for x in desc.Fields]
            # Grab a search cursor
            try:
                sCur = gp.SearchCursor(lineFile, str(whereClause), desc.SpatialReference, "", "")
            except:
                sys.exit("Error opening a SearchCursor on file " + str(lineFile) + ". " + str(sys.exc_info()))
            # Create the output file
            gp.CreateFeatureClass(os.path.dirname(outputFile), os.path.basename(outputFile), "POINT", lineFile, "", "", desc.SpatialReference)
            descOut = gp.Describe(outputFile)
            # Grab an insert cursor
            try:
                iCur = gp.InsertCursor(outputFile, desc.SpatialReference)
            except:
                raise exceptions.CursorError("Error creating Insert Cursor" + str(sys.exc_info()[0]))
            # Start reading the file
            sRow = sCur.Next()
            iRow = iCur.NewRow()
            while sRow:
                # Grab the field values of the row
                valueDict = dict()
                for field in lineFileFields:
                    valueDict[field] = sRow.GetValue(field)
                # Get the geometry
                geom = sRow.GetValue(desc.ShapeFieldName)
                # Check for multipart lines
                if geom.IsMultipart:
                    for partNum in range(geom.PartCount):
                        # Grab a part
                        partArray = geom.GetPart(partNum)
                        # We need to add each part as a point here
                        point = partArray.Next()
                        while point:
                            iRow.SetValue(descOut.ShapeFieldName, point)
                            # Add the attributes
                            for key, value in valueDict.iteritems():
                                if key != desc.ShapeFieldName and key != desc.OIDFieldName:
                                    iRow.SetValue(key, value)
                            # Add the point to the file
                            iCur.InsertRow(iRow)
                            # Delete the temporary row
                            del iRow
                            # Create a new row
                            iRow = iCur.NewRow()
                            # Get the next point
                            point = partArray.Next()
                else:
                    partArray = geom.GetPart(0)
                    point = partArray.Next()
                    while point:
                        iRow.SetValue(descOut.ShapeFieldName, point)
                        for key, value in valueDict.iteritems():
                            if key != desc.ShapeFieldName and key != desc.OIDFieldName: 
                                iRow.SetValue(key, value)
                        iCur.InsertRow(iRow)
                        del iRow
                        iRow = iCur.NewRow()
                        point = partArray.Next()
                sRow = sCur.Next()
        else:
            # File exists but OverwriteOutput is not set. We need to exit gracefully.
            raise IOError("File " + str(outputFile) + " already exists and OverwriteOutput is set to false.")
    else:
        raise IOError("File " + str(pointFile) + " does not exist.")
    # Return
    del iCur
    del sCur
    return outputFile
    
def linesToMultiPoints(lineFile, outputFile, whereClause=None):
    """Convers a line feature class to a multipoint feature class.
    Optional where clause can limit conversion.
    """
    # Begin checking conditions before we load the geoprocessor
    # Does the input file exist?
    if os.path.isfile(lineFile):
        # Is the input file indeed a point feature class?
        # Load the geoprocessor and check
        gp = utils.getGP()
        desc = gp.Describe(lineFile)
        # Check if the DataType is of ShapeFile or Featureclass
        if desc.ShapeType not in utils.constants.polyline:
            raise exceptions.WrongGeometryType("The expected input geometry type was line, not " + desc.ShapeType)
        # Does the output file exist? And can we overwrite it?
        if (os.path.isfile(outputFile) and gp.OverwriteOutput) or not os.path.isfile(outputFile):
            # We can write to outputFile safely. Start reading the input
            if whereClause == None:
                whereClause = ""
            # Find the fields we will need to copy over
            lineFileFields = [x.Name for x in desc.Fields]
            # Grab a search cursor
            try:
                sCur = gp.SearchCursor(lineFile, str(whereClause), desc.SpatialReference, "", "")
            except:
                sys.exit("Error opening a SearchCursor on file " + str(lineFile) + ". " + str(sys.exc_info()))
            # Create the output file
            gp.CreateFeatureClass(os.path.dirname(outputFile), os.path.basename(outputFile), "MULTIPOINT", lineFile, "", "", desc.SpatialReference)
            descOut = gp.Describe(outputFile)
            # Grab an insert cursor
            try:
                iCur = gp.InsertCursor(outputFile, desc.SpatialReference)
            except:
                raise exceptions.CursorError("Error creating Insert Cursor" + str(sys.exc_info()[0]))
            # Start reading the file
            sRow = sCur.Next()
            iRow = iCur.NewRow()
            # Point array
            pntArray = gp.CreateObject("Array")
            while sRow:
                # Grab the field values of the row
                valueDict = dict()
                for field in lineFileFields:
                    valueDict[field] = sRow.GetValue(field)
                # Get the geometry
                geom = sRow.GetValue(desc.ShapeFieldName)
                # Check for multipart lines
                if geom.IsMultipart:
                    for partNum in range(geom.PartCount):
                        # Grab a part
                        partArray = geom.GetPart(partNum)
                        # We need to add each part as a point here
                        point = partArray.Next()
                        while point:
                            pntArray.Add(point)
                            point = partArray.Next()
                else:
                    partArray = geom.GetPart(0)
                    point = partArray.Next()
                    while point:
                        pntArray.Add(point)
                        point = partArray.Next()
                # Add the geometry
                iRow.SetValue(descOut.ShapeFieldName, pntArray)
                # Add the attributes
                for key, value in valueDict.iteritems():
                    if key != desc.ShapeFieldName and key != desc.OIDFieldName:
                        iRow.SetValue(key, value)
                # Save the row
                iCur.InsertRow(iRow)
                # Clear variables
                del iRow
                pntArray.RemoveAll()
                # Create a new row
                iRow = iCur.NewRow()
                sRow = sCur.Next()
        else:
            # File exists but OverwriteOutput is not set. We need to exit gracefully.
            raise IOError("File " + str(outputFile) + " already exists and OverwriteOutput is set to false.")
    else:
        raise IOError("File " + str(lineFile) + " does not exist.")
    # Return
    del iCur
    del sCur
    return outputFile
   
def polygonsToPoints(polygonFile, outputFile, whereClause=None):
    """Converts a polygon feature class to a point feature class.
    Optional where clause can limit conversion.
    """
    # Begin checking conditions before we load the geoprocessor
    # Does the input file exist?
    if os.path.isfile(polygonFile):
        # Is the input file indeed a point feature class?
        # Load the geoprocessor and check
        gp = utils.getGP()
        desc = gp.Describe(polygonFile)
        # Check if the DataType is of ShapeFile or Featureclass
        if desc.ShapeType not in utils.constants.polygon:
            raise exceptions.WrongGeometryType("The expected input geometry type was polygon, not " + desc.ShapeType)
        # Does the output file exist? And can we overwrite it?
        if (os.path.isfile(outputFile) and gp.OverwriteOutput) or not os.path.isfile(outputFile):
            # We can write to outputFile safely. Start reading the input
            if whereClause == None:
                whereClause = ""
            # Find the fields we will need to copy over
            polygonFileFields = [x.Name for x in desc.Fields]
            # Grab a search cursor
            try:
                sCur = gp.SearchCursor(polygonFile, str(whereClause), desc.SpatialReference, "", "")
            except:
                sys.exit("Error opening a SearchCursor on file " + str(polygonFile) + ". " + str(sys.exc_info()))
            # Create the output file
            gp.CreateFeatureClass(os.path.dirname(outputFile), os.path.basename(outputFile), "POINT", polygonFile, "", "", desc.SpatialReference)
            descOut = gp.Describe(outputFile)
            # Grab an insert cursor
            try:
                iCur = gp.InsertCursor(outputFile, desc.SpatialReference)
            except:
                raise exceptions.CursorError("Error creating Insert Cursor" + str(sys.exc_info()[0]))
            # Start reading the file
            sRow = sCur.Next()
            iRow = iCur.NewRow()
            while sRow:
                # Grab the field values of the row
                valueDict = dict()
                for field in polygonFileFields:
                    valueDict[field] = sRow.GetValue(field)
                # Get the geometry
                geom = sRow.GetValue(desc.ShapeFieldName)
                # Check for multipart lines
                if geom.IsMultipart:
                    for partNum in range(geom.PartCount):
                        # Grab a part
                        partArray = geom.GetPart(partNum)
                        # We need to add each part as a point here
                        point = partArray.Next()
                        while point:
                            iRow.SetValue(descOut.ShapeFieldName, point)
                            # Add the attributes
                            for key, value in valueDict.iteritems():
                                if key != desc.ShapeFieldName and key != desc.OIDFieldName:
                                    iRow.SetValue(key, value)
                            # Add the point to the file
                            iCur.InsertRow(iRow)
                            # Delete the temporary row
                            del iRow
                            # Create a new row
                            iRow = iCur.NewRow()
                            # Get the next point
                            point = partArray.Next()
                else:
                    partArray = geom.GetPart(0)
                    point = partArray.Next()
                    while point:
                        iRow.SetValue(descOut.ShapeFieldName, point)
                        for key, value in valueDict.iteritems():
                            if key != desc.ShapeFieldName and key != desc.OIDFieldName: 
                                iRow.SetValue(key, value)
                        iCur.InsertRow(iRow)
                        del iRow
                        iRow = iCur.NewRow()
                        point = partArray.Next()
                sRow = sCur.Next()
        else:
            # File exists but OverwriteOutput is not set. We need to exit gracefully.
            raise IOError("File " + str(outputFile) + " already exists and OverwriteOutput is set to false.")
    else:
        raise IOError("File " + str(polygonFile) + " does not exist.")
    # Return
    del iCur
    del sCur
    return outputFile
    
def polygonsToMultiPoints(polygonFile, outputFile, whereClause=None):
    """Converts a polygon feature class to a multipoint feature class.
    Optional where clause can limit conversion.
    """
    # Begin checking conditions before we load the geoprocessor
    # Does the input file exist?
    if os.path.isfile(polygonFile):
        # Is the input file indeed a point feature class?
        # Load the geoprocessor and check
        gp = utils.getGP()
        desc = gp.Describe(polygonFile)
        # Check if the DataType is of ShapeFile or Featureclass
        if desc.ShapeType not in utils.constants.polygon:
            raise exceptions.WrongGeometryType("The expected input geometry type was polygon, not " + desc.ShapeType)
        # Does the output file exist? And can we overwrite it?
        if (os.path.isfile(outputFile) and gp.OverwriteOutput) or not os.path.isfile(outputFile):
            # We can write to outputFile safely. Start reading the input
            if whereClause == None:
                whereClause = ""
            # Find the fields we will need to copy over
            polygonFileFields = [x.Name for x in desc.Fields]
            # Grab a search cursor
            try:
                sCur = gp.SearchCursor(polygonFile, str(whereClause), desc.SpatialReference, "", "")
            except:
                sys.exit("Error opening a SearchCursor on file " + str(lineFile) + ". " + str(sys.exc_info()))
            # Create the output file
            gp.CreateFeatureClass(os.path.dirname(outputFile), os.path.basename(outputFile), "MULTIPOINT", polygonFile, "", "", desc.SpatialReference)
            descOut = gp.Describe(outputFile)
            # Grab an insert cursor
            try:
                iCur = gp.InsertCursor(outputFile, desc.SpatialReference)
            except:
                raise exceptions.CursorError("Error creating Insert Cursor" + str(sys.exc_info()[0]))
            # Start reading the file
            sRow = sCur.Next()
            iRow = iCur.NewRow()
            # Point array
            pntArray = gp.CreateObject("Array")
            while sRow:
                # Grab the field values of the row
                valueDict = dict()
                for field in polygonFileFields:
                    valueDict[field] = sRow.GetValue(field)
                # Get the geometry
                geom = sRow.GetValue(desc.ShapeFieldName)
                # Check for multipart lines
                if geom.IsMultipart:
                    for partNum in range(geom.PartCount):
                        # Grab a part
                        partArray = geom.GetPart(partNum)
                        # We need to add each part as a point here
                        point = partArray.Next()
                        while point:
                            pntArray.Add(point)
                            point = partArray.Next()
                else:
                    partArray = geom.GetPart(0)
                    point = partArray.Next()
                    while point:
                        pntArray.Add(point)
                        point = partArray.Next()
                # Add the geometry
                iRow.SetValue(descOut.ShapeFieldName, pntArray)
                # Add the attributes
                for key, value in valueDict.iteritems():
                    if key != desc.ShapeFieldName and key != desc.OIDFieldName:
                        iRow.SetValue(key, value)
                # Save the row
                iCur.InsertRow(iRow)
                # Clear variables
                del iRow
                pntArray.RemoveAll()
                # Create a new row
                iRow = iCur.NewRow()
                sRow = sCur.Next()
        else:
            # File exists but OverwriteOutput is not set. We need to exit gracefully.
            raise IOError("File " + str(outputFile) + " already exists and OverwriteOutput is set to false.")
    else:
        raise IOError("File " + str(lineFile) + " does not exist.")
    # Return
    del iCur
    del sCur
    return outputFile

def kmlToFC(kmlFile, outputFile):
    """Convert a KML file into a shapefile. If there are multiple
    geometries in the KML file, multiple shapefiles will be created
    based on the geometries. Note, it only handles Points, LineStrings
    and Polygons now.
    """
    # Check if the KML file exists
    if os.path.isfile(kmlFile):
        # Create a geoprocessor object
        gp = utils.getGP()
        # Check if we can write the output
        if (os.path.isfile(outputFile) and gp.OverwriteOutput) or not os.path.isfile(outputFile):
            # Variable to help us figure out what feature class to create
            geomType = []
            # Placemark counter
            placemarkNum = 0
            # Create a Placemark Class to store the placemarks
            class Placemark():
                def __init__(self):
                    self.name = ""
                    self.description = ""
                    self.coordinates = []
                    self.type = ""
                    self.partNum = 0
                def __str__(self):
                    return self.name + "\n" + "\t" + str(self.description) + "\n" + "\t" + str(self.type)+ "\n" +  "\t\t" + str(self.coordinates)
            # Create a method to extract coordinate
            def extractCoords(node):
                for child in node.childNodes:
                    if child.nodeName == 'coordinates':
                        # Grab the coordinate children
                        for coordChild in child.childNodes:
                            if coordChild.nodeType == coordChild.TEXT_NODE:
                                return coordChild.wholeText
            # Grab the geometry and return it as a placemark
            def getGeometry(node, feat):
                if child.nodeName == 'Point':
                    feat.type = 'POINT'
                    coordsString = extractCoords(child)
                    for pnt in coordsString.split():
                        coords = pnt.split(',')
                        feat.coordinates.append((float(coords[0]), float(coords[1]))) 
                    if 'POINT' not in geomType:
                        geomType.append('POINT')
                    feat.partNum = 1
                elif child.nodeName == 'LineString':
                    feat.type = 'POLYLINE'
                    coordsString = extractCoords(child)
                    for pnt in coordsString.split():
                        coords = pnt.split(',')
                        feat.coordinates.append((float(coords[0]), float(coords[1])))
                    if 'POLYLINE' not in geomType:
                        geomType.append('POLYLINE')
                    feat.partNum = 1
                elif child.nodeName == 'Polygon':
                    feat.type = 'POLYGON'
                    # Get the outer boundary of the polygon
                    outBnd = child.getElementsByTagName('outerBoundaryIs')
                    # Get the linear ring in the outer boundary
                    linRing = outBnd[0].getElementsByTagName('LinearRing')
                    coordsString = extractCoords(linRing.item(0))
                    outerRing = []
                    for pnt in coordsString.split():
                        coords = pnt.split(',')
                        outerRing.append((float(coords[0]), float(coords[1])))
                    # Check for polygon rings
                    try:
                        ringBnds = child.getElementsByTagName('innerBoundaryIs')
                    except:
                        ringBnds = []
                    # If we have polygon rings, read them
                    if len(ringBnds) > 0:
                        feat.coordinates.append(None)
                        # Add the outer ring
                        feat.coordinates.append(outerRing)
                        feat.partNum += 1
                        # Loop through inner rings
                        for ring in ringBnds:
                            ringRing = ring.getElementsByTagName('LinearRing')
                            # Get the coordinates of ring
                            coordsString = extractCoords(ringRing.item(0))
                            innerRing = []
                            for pnt in coordsString.split():
                                coords = pnt.split(',')
                                innerRing.append((float(coords[0]), float(coords[1])))
                            # Add a None object between outer and inner rings
                            feat.coordinates.append(None)
                            # Add the inner ring
                            feat.coordinates.append(innerRing)
                            del innerRing
                            feat.partNum += 1
                    if 'POLYGON' not in geomType:
                        geomType.append('POLYGON')
                return feat
            # Parse the KML file
            kmlDom = minidom.parse(kmlFile)
            # Empty list of all features we found
            featList = []
            # Go through all the Placemarks (features)
            for node in kmlDom.getElementsByTagName('Placemark'):
                feat = Placemark()
                # Go through the node children to find the records
                for child in node.childNodes:
                    # Check the child geometry type
                    if child.nodeName == 'Point' or child.nodeName == 'LineString' or child.nodeName == 'Polygon':
                        feat = getGeometry(child, feat)
                    #
                    # IMPLEMENT THE MULTIGEOMETRY PARSER
                    #
                    #elif child.nodeName == 'MultiGeometry':
                        # Multigeometries are treated as a collection of separate features
                        #feat.id = placemarkNum
                    # Grab the name
                    if child.nodeName == 'name':
                        for subChild in child.childNodes:
                            if subChild.nodeType == subChild.TEXT_NODE:
                                feat.name = subChild.wholeText
                    # Grab the description
                    if child.nodeName == 'description':
                        for subChild in child.childNodes:
                            if subChild.nodeType == subChild.TEXT_NODE or subChild.nodeType == subChild.CDATA_SECTION_NODE:
                                descrText = subChild.wholeText
                                if "<![CDATA[" in descrText:
                                    feat.description = descrText[11:-3]
                                else:
                                    feat.description = descrText
                    # Increment the placemark ID
                    placemarkNum += 1                                
                # Add our feature to the list
                featList.append(feat)
                del feat
            # Now we have a list of all features, time to add them
            # Create the output feature classes as needed
            for fcType in geomType:
                # If only one geometry present, use default output name
                outFile = []
                # (4326 is the code for WGS84 coordinate system
                if len(geomType) == 1:
                    gp.CreateFeatureClass(os.path.dirname(outputFile), os.path.basename(outputFile), str(fcType), "", "", "", "4326")
                    outFile.append(dict({fcType: outputFile}))
                else:
                    # We have multiple geographies. Check if we are using shapefiles
                    # or geodatabase here.
                    if os.path.isdir(os.path.dirname(outputFile)): 
                        gp.CreateFeatureClass(os.path.dirname(outputFile), os.path.basename(outputFile)[:4] + str(fcType) + ".shp", fcType, "", "", "", "4326")
                        outFile.append(dict({fcType: outputFile[:4] + fcType + ".shp"}))
                    else:
                        # The path is not a directory, so assume it is a geodatabase
                        gp.CreateFeatureClass(os.path.dirname(outputFile), os.path.basename(outputFile) + str(fcType), fcType, "", "", "", "4326")
                        outFile.append(dict({fcType: outputFile + fcType}))
            # Add to each feature class the "Name" attribute and description
            for file in outFile:
                # Find the maximum length we need
                nameLength = max([len(feat.name) for feat in featList])
                try:
                    gp.AddField(file.values()[0], "Name", "TEXT", "", "", nameLength)
                except:
                    raise IOError("Cannot add field. Is the outfile in use?")
                descLength = max([len(feat.description) for feat in featList])
                try:
                    gp.AddField(file.values()[0], "Descr", "TEXT", "", "", descLength)
                except:
                    raise IOError("Cannot add field. Is the outfile in use?")
            # Now that we have the feature classes, start saving data
            # Create the insert cursors
            iCurs = dict()
            for file in outFile:
                # File is a dict with key the feature type and value the filename
                iCurs[file.keys()[0]] = gp.InsertCursor(file.values()[0], "4326")
            # Loop through the features and add them
            for feat in featList:
                if feat.type == "POINT":
                    # Create empty point object
                    pnt = gp.CreateObject("Point")
                    # Populate point with our values
                    pnt.x, pnt.y = feat.coordinates[0]
                    # Create a new row
                    iRow = iCurs[feat.type].NewRow()
                    # Set the geometry to the point
                    iRow.shape = pnt
                    # Set the name
                    iRow.SetValue("Name", feat.name)
                    # Set the description
                    iRow.SetValue("Descr", feat.description)
                    # Insert the row
                    iCurs[feat.type].InsertRow(iRow)
                    # Delete the objects
                    del pnt
                    del iRow
                elif feat.type == "LINE" or feat.type == "POLYGON":
                    # Create empty point object
                    pnt = gp.CreateObject("Point")
                    # Create empty array object
                    shpArray = gp.CreateObject("Array")
                    prtArray = gp.CreateObject("Array")
                    # Start grabbing points
                    if feat.partNum > 1:
                        # It's multipart. Start reading parts
                        partCount = 0
                        for part in feat.coordinates:
                            # If the part is None, it is a separator.
                            if part != None:
                                for point in part:
                                    pnt.x, pnt.y = point
                                    prtArray.Add(pnt)
                            else:
                                shpArray.add(prtArray)
                                prtArray.RemoveAll()
                                # Add an empty point to signal change in rings
                                #prtArray.Add(emptyPnt)
                            partCount += 1
                        shpArray.Add(prtArray)
                    else:
                        for part in feat.coordinates:
                            pnt.x, pnt.y = part
                            # Add the point to the array
                            shpArray.Add(pnt)
                    # Now that the array is full, add it
                    # Create a new row
                    iRow = iCurs[feat.type].NewRow()
                    # Set the geometry to the array
                    iRow.shape = shpArray
                    # Set the name and description
                    iRow.SetValue("Name", feat.name)
                    iRow.SetValue("Descr", feat.description)
                    # Insert the row
                    iCurs[feat.type].InsertRow(iRow)
                    # Delete the objects
                    del pnt
                    shpArray.RemoveAll
                    prtArray.RemoveAll
                    del iRow
                del shpArray
        else:
            raise IOError("Cannot write to output file.")
    else:
        raise IOError("File " + str(kmlFile) + " does not exist.")
    if len(outFile) == 1:
        return outFile[0]
    else:
        return (x for x in outFile)

def fcToKML(fcFile, outputFile, whereClause=None):
    """Export a feature class to a KML file. Optionally use a
    where clause to limit the features converted.
    """
    # Check if the KML file exists
    if os.path.isfile(fcFile):
        # Create a geoprocessor object
        gp = utils.getGP()
        # Check if we can write the output
        if (os.path.isfile(outputFile) and gp.OverwriteOutput) or not os.path.isfile(outputFile):
            # Create the output file
            kmlDoc = minidom.Document()
            # Write the header information
            kmlElement = kmlDoc.createElementNS('http://earth.google.com/kml/2.2', 'kml')
            kmlElement.setAttribute('xmlns','http://earth.google.com/kml/2.2')
            kmlElement = kmlDoc.appendChild(kmlElement)
            documentElement = kmlDoc.createElement('Document')
            documentElement = kmlElement.appendChild(documentElement)
            # Get information about the feature class
            desc = gp.Describe(fcFile)
            geoType = desc.ShapeType
            # Get the names of our fields
            fcFields = [x.Name for x in desc.Fields]
            # Check if we have a where clause
            if whereClause == None:
                whereClause = ""
            # Create our search cursor for the file, use the WGS84 coordinate system
            try:
                sCur = gp.SearchCursor(fcFile, whereClause, "4326")
            except:
                raise exceptions.CursorError(str(sys.exc_info()[0]))
            # Loop through our fc
            sRow = sCur.Next()
            while sRow:
                # Get the Geometry
                geom = sRow.GetValue(desc.ShapeFieldName)
                # Create a placemark
                placemark = kmlDoc.createElement('Placemark')
                # Create extended data
                extData = kmlDoc.createElement('ExtendedData')
                # Append the extData to Placemark
                placemark.appendChild(extData)
                # Create the data element
                #dataEl = kmlDoc.createElement('Data')
                for fieldName in fcFields:
                    if fieldName != desc.ShapeFieldName:
                        # Create the data element
                        dataEl = kmlDoc.createElement('Data')
                        dataEl.setAttribute('name', fieldName)
                        valEl = kmlDoc.createElement('value')
                        dataEl.appendChild(valEl)
                        valText = kmlDoc.createTextNode(str(sRow.GetValue(fieldName)))
                        valEl.appendChild(valText)
                        extData.appendChild(dataEl)
                if geoType in utils.constants.point:
                    point = geom.GetPart()
                    pointEl = kmlDoc.createElement('Point')
                    placemark.appendChild(pointEl)
                    coorEl = kmlDoc.createElement('coordinates')
                    coorText = kmlDoc.createTextNode(str(point.x) + "," + str(point.y))
                    coorEl.appendChild(coorText)
                    pointEl.appendChild(coorEl)
                    documentElement.appendChild(placemark)
                elif geoType in utils.constants.multipoint:
                    for part in range(geom.PartCount):
                        point = geom.GetPart(part)
                        pointEl = kmlDoc.createElement('Point')
                        placemark.appendChild(pointEl)
                        coorEl = kmlDoc.createElement('coordinates')
                        coorText = kmlDoc.createTextNode(str(point.x) + "," + str(point.y))
                        coorEl.appendChild(coorText)
                        pointEl.appendChild(coorEl)
                    documentElement.appendChild(placemark)
                elif geoType in utils.constants.polyline:
                    for part in range(geom.PartCount):
                        lineArray = geom.GetPart(part)
                        coords = ""
                        point = lineArray.Next()
                        while point:
                            coords += str(point.x) + "," + str(point.y) + " "
                            point = lineArray.Next()
                        lineEl = kmlDoc.createElement('LinearString')
                        placemark.appendChild(lineEl)
                        coorEl = kmlDoc.createElement('coordinates')
                        coorText = kmlDoc.createTextNode(coords)
                        coorEl.appendChild(coorText)
                        lineEl.appendChild(coorEl)
                    documentElement.appendChild(placemark)
                elif geoType in utils.constants.polygon:
                    pass
                sRow = sCur.Next()
            kmlFile = open(outputFile, 'w')
            kmlFile.write(kmlDoc.toprettyxml('    ', newl= '\n', encoding='utf-8'))
            kmlFile.close()
        else:
            raise IOError("Cannot write to output file.")
    else:
        raise IOError("File " + str(fcFile) + " does not exist.")
    return outputFile
    
    




