"""Table conversion methods, part of gptools.
"""

# Import all necessary modules
import os                  # Could be security risk since file is not locked
import sys
import math

import xlrd
from dbfpy.dbf import *
from dbfpy.dbfnew import *
from dbfpy.record import *



import gptools.utils as utils
import gptools.utils.exceptions as exceptions

def xlsToDBF(xlsFile, dbfFile, workSheet=None, header=True):
    """Convert an Excel File to a dbfFile. If worksheet name is not given,
    the first one will be used.
    """
    if os.path.isfile(xlsFile):
        # Open the XLS file using xlrd
        try:
            wBook = xlrd.open_workbook(xlsFile, formatting_info=True)
        except:
            sys.exit("Could not open Excel File " + str(xlsFile) + ". Exiting...")
        # Check if we are opening a specific worksheet
        if workSheet != None:
            # Does the worksheet name exist?
            if workSheet in wBook.sheet_names():
                # Open the worksheet
                wSheet = wBook.sheet_by_name(workSheet)
            else:
                # A worksheet with that name does not exist. Exit
                sys.exit("No sheet named " + str(workSheet) + " found. Exiting...")
        else:
            # Check if there are any work sheets
            if wBook.nsheets:
                # Open the first worksheet
                wSheet = wBook.sheet_by_index(0)
        # Did the user tell us there are headers?
        if header:
            # The user indicated first row is headers. Read it
            headerNames = wSheet.row_values(0, start_colx=0, end_colx=wSheet.ncols)
        else:
            # There were no headers. Just use the alphabet
            if wSheet.nrows <= 26:
                headerNames = [xlrd.cellname(0, x)[0] for x in range(wSheet.nrows)]
            else:
                headerNames = [xlrd.cellname(0, x)[:2] for x in range(wSheet.nrows)]
        # Identify the type of data we have. This will come from the data in the
        #   first row. Relies of properly formed XLS sheets
        dataTypes = wSheet.row_types(1, start_colx=0, end_colx=wSheet.ncols)
        # Find the size of each column. Depends on proper formating of file
        colSizes = [(wSheet.computed_column_width(x) /256) for x in range(wSheet.ncols)]
        # Make sure we got it all right
        if len(headerNames) != len(dataTypes) != len(colSizes):
            sys.exit("We got a problem reading the file correctly")
        # Create the DBF file with headers and types as given
        if os.path.isfile(dbfFile):
            raise IOError("File " + str(dbfFile) + " already exists.")
        else:
            # Creare a new DBF
            newDbf = dbf_new()
            # Iterate over header names and types
            for x in range(len(headerNames)):
                # 0 and 1 indicate strings
                if dataTypes[x] == 0 or dataTypes[x] == 1:
                    newDbf.add_field(headerNames[x], 'C', colSizes[x])
                # 2 indicates numerical
                elif dataTypes[x] == 2:
                    # There is no way to understand if Excel numerical is float or int
                    # Assume float with decimals = half the col size
                    newDbf.add_field(headerNames[x], 'N', math.ceil(colSizes[x]), math.ceil(colSizes[x] / 2))
                # 3 indicates date - NOT IMPLEMENTED YET
            # Save the DBF file
            newDbf.write(dbfFile) 
        # Open our dbf file for writing now
        newDbf = Dbf(dbfFile, False)        
        # Begin reading data then
        for rowNum in range(wSheet.nrows):
            # Create a new dbf row
            dbfRow = DbfRecord(newDbf)
            if header and rowNum == 0:
                pass
            else:
                # Read the whole row from XLS
                rowValues = wSheet.row_values(rowNum, start_colx=0, end_colx=wSheet.ncols)
                # Write each value to the DBF record
                for x in range(len(rowValues)):
                    dbfRow[headerNames[x]] = rowValues[x]
                # Save the record
                dbfRow.store()
        # Close the DBF file
        newDbf.close()
        # Close the EXCEL file
        del wSheet
        del wBook
    else:
        raise IOError("File " + str(xlsFile) + " does not exist.")
    return dbfFile